\echo Use "CREATE EXTENSION vis" to load this file. \quit
/*
 * Main pgVis definition type
 */

CREATE DOMAIN @extschema@.t_definition AS JSONB
CONSTRAINT valid_definition CHECK (VALUE->'pgvis'->>'version' IS NOT NULL);
/*
 * A dataset which will be visualised
 */

CREATE DOMAIN @extschema@.t_dataset AS JSONB
CONSTRAINT valid_dataset CHECK (VALUE->>'_type' = 'dataset');
/*
 * A tile on a visulisation dashboard
 */

CREATE DOMAIN @extschema@.t_tile AS JSONB
CONSTRAINT valid_tile CHECK (VALUE->>'_type' = ANY (ARRAY['chart', 'data_table', 'tile']));
/*
 * A set of options. a type mostly used to offer a nice DSL
 */

CREATE DOMAIN @extschema@.t_options AS JSONB
CONSTRAINT valid_options CHECK (VALUE->>'_type' = 'options');
/*
 * A column of a data table
 */

CREATE DOMAIN @extschema@.t_column AS JSONB
CONSTRAINT valid_column CHECK (VALUE->>'_type' = 'column');
/*
 * A series on a chart
 */

CREATE DOMAIN @extschema@.t_series AS JSONB
CONSTRAINT valid_series CHECK (VALUE->>'_type' = 'series');
/*
 * A simple aggregate which will shallow merge JSONB objects together, a little like JS Object.assign
 */

CREATE OR REPLACE FUNCTION @extschema@.jsonb_merge_agg_accumulate(st JSONB, val JSONB)
RETURNS JSONB LANGUAGE SQL AS $$
SELECT st || coalesce(val, '{}'::JSONB);
$$;


CREATE OR REPLACE AGGREGATE @extschema@.jsonb_merge_agg(JSONB)
(
    SFUNC = @extschema@.jsonb_merge_agg_accumulate,
    STYPE = jsonb,
    INITCOND = '{}'
);
/*
 * pgVis Options DSL
 */

CREATE OR REPLACE FUNCTION @extschema@.option(n TEXT, v TEXT)
RETURNS @extschema@.t_options LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'options', n, v);
$$;

CREATE OR REPLACE FUNCTION @extschema@.option(n TEXT, v NUMERIC)
RETURNS @extschema@.t_options LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'options', n, v);
$$;

CREATE OR REPLACE FUNCTION @extschema@.option(n TEXT, v BOOLEAN)
RETURNS @extschema@.t_options LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'options', n, v);
$$;

CREATE OR REPLACE FUNCTION @extschema@.option(n TEXT, v VARIADIC JSONB[])
RETURNS @extschema@.t_options LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'options', n, (SELECT @extschema@.jsonb_merge_agg(p) - '_type' FROM unnest(v) p));
$$;

/*
 * pgVis dataset DSL
 */

CREATE OR REPLACE FUNCTION @extschema@.dataset_autodescribe_labels(row_obj JSONB)
RETURNS JSONB LANGUAGE SQL AS $$
SELECT jsonb_object_agg(k.key, jsonb_build_object('summary', q.description))
FROM jsonb_object_keys(row_obj) AS k(key)
LEFT JOIN (
    SELECT a.attname, d.description
    FROM pg_attribute a
    JOIN pg_description d ON (a.attrelid = d.objoid AND a.attnum = d.objsubid)
) q ON (q.attname = k.key);
$$;


CREATE OR REPLACE FUNCTION @extschema@.labels(pairs VARIADIC TEXT[])
RETURNS JSONB LANGUAGE SQL AS $$
SELECT jsonb_object_agg(f.k, jsonb_build_object('summary', f.v)) 
FROM jsonb_each(jsonb_build_object(VARIADIC pairs)) f(k,v);
$$;


CREATE OR REPLACE FUNCTION @extschema@.dataset(
    data JSONB,
    labels JSONB DEFAULT NULL
)
RETURNS @extschema@.t_dataset LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'dataset', 'data', COALESCE(data, '[]'::JSONB), 'labels', COALESCE(labels, @extschema@.dataset_autodescribe_labels(data->0), '{}'::JSONB));
$$;


CREATE OR REPLACE FUNCTION @extschema@.dataset(id TEXT)
RETURNS @extschema@.t_dataset LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'dataset', 'id', id);
$$;


CREATE OR REPLACE FUNCTION @extschema@.dataset(
    id TEXT,
    summary TEXT,
    data JSONB,
    labels JSONB DEFAULT NULL
)
RETURNS @extschema@.t_dataset LANGUAGE SQL AS $$
SELECT @extschema@.dataset(id) || jsonb_build_object('summary', summary, 'data', COALESCE(data, '[]'::JSONB), 'labels', COALESCE(labels, @extschema@.dataset_autodescribe_labels(data->0), '{}'::JSONB));
$$;
/*
 * pgVis base tile definition
 */

CREATE OR REPLACE FUNCTION @extschema@.tile(
    _type TEXT, 
    id TEXT, 
    size INTEGER, 
    summary TEXT, 
    props JSONB
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT coalesce(props, '{}'::JSONB) || jsonb_build_object(
    '_type', _type,
    'id', id, 
    'summary', summary,
    'size', size
);
$$;
/*
 * pgVis chart series DSL
 */

CREATE OR REPLACE FUNCTION @extschema@.series(summary TEXT, x TEXT, y TEXT, options JSONB DEFAULT '{}')
RETURNS @extschema@.t_series LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'series', 'summary', summary, 'x', x, 'y', y, 'options', options);
$$;
/*
 * pgVis chart tile DSL
 */

CREATE OR REPLACE FUNCTION @extschema@.chart(
    style TEXT, 
    id TEXT,
    size INTEGER,
    summary TEXT,
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.tile('chart', id, size, summary, jsonb_build_object(
    'style', style,
    'dataset', (SELECT p - '_type' FROM unnest(parts) p WHERE p->>'_type' = 'dataset' LIMIT 1),
    'series', (SELECT jsonb_agg(p - '_type') FROM unnest(parts) p WHERE p->>'_type' = 'series'),
    'options', (SELECT @extschema@.jsonb_merge_agg(p) - '_type' FROM unnest(parts) p WHERE p->>'_type' = 'options')
));
$$;



CREATE OR REPLACE FUNCTION @extschema@.bar_chart(
    id TEXT,
    size INTEGER,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.chart('bar', id, size, summary, VARIADIC parts);
$$;


CREATE OR REPLACE FUNCTION @extschema@.bar_chart(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.bar_chart(null, 12, summary, VARIADIC parts);
$$;



CREATE OR REPLACE FUNCTION @extschema@.line_chart(
    id TEXT,
    size INTEGER,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.chart('line', id, size, summary, VARIADIC parts);
$$;


CREATE OR REPLACE FUNCTION @extschema@.line_chart(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.line_chart(null, 12, summary, VARIADIC parts);
$$;



CREATE OR REPLACE FUNCTION @extschema@.scatter_chart(
    id TEXT,
    size INTEGER,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.chart('scatter', id, size, summary, VARIADIC parts);
$$;


CREATE OR REPLACE FUNCTION @extschema@.scatter_chart(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.scatter_chart(null, 12, summary, VARIADIC parts);
$$;



CREATE OR REPLACE FUNCTION @extschema@.pie_chart(
    id TEXT,
    size INTEGER,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.chart('pie', id, size, summary, VARIADIC parts);
$$;


CREATE OR REPLACE FUNCTION @extschema@.pie_chart(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.pie_chart(null, 12, summary, VARIADIC parts);
$$;



CREATE OR REPLACE FUNCTION @extschema@.polar_area_chart(
    id TEXT,
    size INTEGER,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.chart('polarArea', id, size, summary, VARIADIC parts);
$$;


CREATE OR REPLACE FUNCTION @extschema@.polar_area_chart(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.polar_area_chart(null, 12, summary, VARIADIC parts);
$$;



CREATE OR REPLACE FUNCTION @extschema@.radar_chart(
    id TEXT,
    size INTEGER,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.chart('radar', id, size, summary, VARIADIC parts);
$$;


CREATE OR REPLACE FUNCTION @extschema@.radar_chart(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.radar_chart(null, 12, summary, VARIADIC parts);
$$;
/*
 * pgVis data table column DSL
 */

CREATE OR REPLACE FUNCTION @extschema@.column(summary TEXT, colName TEXT, options JSONB DEFAULT '{}')
RETURNS @extschema@.t_column LANGUAGE SQL AS $$
SELECT jsonb_build_object('_type', 'column', 'summary', summary, 'column', colName, 'options', options);
$$;
/*
 * pgVis data table tile DSL
 */

CREATE OR REPLACE FUNCTION @extschema@.data_table(
    id TEXT,
    size INTEGER,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.tile('data_table', id, size, summary, jsonb_build_object(
    'dataset', (SELECT p - '_type' FROM unnest(parts) p WHERE p->>'_type' = 'dataset' LIMIT 1),
    'columns', (SELECT jsonb_agg(p - '_type') FROM unnest(parts) p WHERE p->>'_type' = 'column'),
    'options', (SELECT @extschema@.jsonb_merge_agg(p) - '_type' FROM unnest(parts) p WHERE p->>'_type' = 'options')
));
$$;


CREATE OR REPLACE FUNCTION @extschema@.data_table(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_tile LANGUAGE SQL AS $$
SELECT @extschema@.data_table(null, 12, summary, VARIADIC parts);
$$;
/*
 * pgVis main visualisation DSL
 */
 
CREATE OR REPLACE FUNCTION @extschema@.visualise(
    id TEXT,
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT jsonb_build_object(
    'pgvis', jsonb_build_object('version', 1),
    'id', coalesce(id, regexp_replace(lower(summary), '[^a-z0-9]', '', 'g'), gen_random_uuid()::TEXT),
    'summary', summary,
    'datasets', (SELECT @extschema@.jsonb_merge_agg(jsonb_build_object(coalesce(p.v->>'id', 'dataset-' || p.i ), p.v - '_type')) FROM unnest(parts) WITH ORDINALITY p(v,i) WHERE p.v->>'_type' = 'dataset'),
    'tiles', (SELECT jsonb_agg(p.v || jsonb_build_object('id', coalesce(p.v->>'id', 'tile-' || p.i))) FROM unnest(parts) WITH ORDINALITY p(v,i) WHERE p.v->>'_type' = ANY (ARRAY['chart', 'data_table', 'tile'])),
    'options', (SELECT @extschema@.jsonb_merge_agg(p) - '_type' FROM unnest(parts) p WHERE p->>'_type' = 'options')
);
$$;


CREATE OR REPLACE FUNCTION @extschema@.visualise(
    summary TEXT, 
    parts VARIADIC JSONB[]
)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT @extschema@.visualise(null, summary, VARIADIC parts);
$$;
/*
 * pgVis HTML output formats
 */

CREATE OR REPLACE FUNCTION @extschema@.html(definition @extschema@.t_definition)
RETURNS TEXT LANGUAGE SQL AS $$
SELECT 
$h$
<html>
    <head>
        <title>pgVis</title>
        <style> body { font-size: 16px; background-color: #fff; color: #2c2c2c; } </style>
        <link rel="stylesheet" href="https://pgvis.org/latest/pgvis-renderer/pgvis.css"/>
        <script src="https://pgvis.org/latest/pgvis-renderer/pgvis-renderer.umd.min.js"></script>
    </head>
    <body>
        <div id="pgvis"></div>
        <script>pgVis(
$h$ ||
jsonb_pretty(definition) ||
$h$
          , {"container": "pgvis", "updatePageTitle": true});
        </script>
    </body>
</html>
$h$;
$$;

COMMENT ON FUNCTION @extschema@.html(definition @extschema@.t_definition) IS $$
Wrap the given pgVis visualisation into a single HTML page, ready to be viewed in a browser.
$$;
/*
 * pgVis stored visualisation helpers
 */
 
CREATE OR REPLACE VIEW @extschema@.stored_visualisations AS 
SELECT DISTINCT ON (pn.nspname, p.proname)
    (pn.nspname || '.' || p.proname) AS id, 
    pn.nspname AS schema,
    p.proname AS name,
    p.pronargs AS arity,
    (SELECT COALESCE(jsonb_agg(jsonb_build_object('name', a.n, 'type', (COALESCE(NULLIF(n.nspname || '.', 'pg_catalog.'), '') || t.typname))), '[]'::JSONB)
        FROM unnest(p.proargnames, p.proargtypes) a(n,t)
        JOIN pg_type t ON (a.t = t.oid)
        JOIN pg_namespace n ON (t.typnamespace = n.oid)) AS arguments,
    initcap(regexp_replace(p.proname, '[_]', ' ', 'g')) AS summary,
    regexp_replace(d.description, '(\A\n)|(\n\Z)', '', 'g') AS description,
    has_function_privilege(current_role, p.oid, 'execute') AS can_execute
FROM pg_proc p
JOIN pg_namespace pn ON (p.pronamespace = pn.oid)
JOIN pg_type prt ON (p.prorettype = prt.oid)
JOIN pg_namespace prtn ON (prt.typnamespace = prtn.oid)
LEFT JOIN pg_description d ON (p.oid = d.objoid)
WHERE p.prokind = 'f'
AND (prt.typtype = 'd' AND prt.typname = 't_definition' AND prtn.nspname = '@extschema@')
AND pn.nspname != '@extschema@'
ORDER BY pn.nspname, p.proname;



CREATE OR REPLACE FUNCTION @extschema@.list_stored_visualisations()
RETURNS JSONB LANGUAGE SQL AS $$
SELECT jsonb_agg(to_jsonb(v.*)) FROM @extschema@.stored_visualisations v;
$$;

COMMENT ON FUNCTION @extschema@.list_stored_visualisations() IS $$
List all stored visulations which can be found in the current database
$$;



CREATE OR REPLACE FUNCTION @extschema@.get_stored_visualisations(id TEXT)
RETURNS JSONB LANGUAGE SQL AS $$
SELECT to_jsonb(q.*)
FROM (
    SELECT v.* 
    FROM @extschema@.stored_visualisations v
    WHERE v.schema = (regexp_split_to_array(id, '[.]'))[1]
    AND v.name = (regexp_split_to_array(id, '[.]'))[2]
    LIMIT 1
) q;
$$;

COMMENT ON FUNCTION @extschema@.get_stored_visualisations(id TEXT) IS $$
Lookup a specific stored visulation which can be found in the current database
$$;



CREATE OR REPLACE FUNCTION @extschema@.run_stored_visualisation(id TEXT, args JSONB DEFAULT NULL)
RETURNS SETOF vis.t_definition LANGUAGE PLPGSQL AS $$
DECLARE
    v_info JSONB;
    v_args TEXT;
    v_run TEXT;
BEGIN
    args := COALESCE(args, '{}'::JSONB);
    /* Lookup the visualisation */
    v_info := @extschema@.get_stored_visualisations(id);
    IF (v_info IS NULL) THEN
        RAISE EXCEPTION 'No such visualisation %', id;
    END IF;
    IF (NOT (v_info->>'can_execute')::BOOLEAN) THEN
        RAISE EXCEPTION 'The current user cannot execute the visualisation %', id;
    END IF;
    /* Build the arguments mappings */
    SELECT COALESCE(string_agg(CASE WHEN lower(a->>'type') = 'jsonb' THEN '$1::JSONB' ELSE '$1->>' || quote_literal(a->>'name') || '::' || (a->>'type') END, ', '), '') INTO v_args
    FROM jsonb_array_elements(v_info->'arguments') a;
    /* Build the runtime query */
    v_run := 'SELECT (' || quote_ident(v_info->>'schema') || '.' || quote_ident(v_info->>'name') || '(' || v_args || ') || jsonb_build_object(''id'',' || quote_literal(v_info->>'id')  || '))::@extschema@.t_definition ;';
    /* Execute */
    RETURN QUERY EXECUTE v_run USING args;
END;
$$;

COMMENT ON FUNCTION @extschema@.run_stored_visualisation(id TEXT, args JSONB) IS $$
Execute the stored visualisation with the given id and arguments
$$;
/*
 * pgVis quick util functions
 */

CREATE OR REPLACE FUNCTION @extschema@.quick_bar_chart(summary TEXT, x TEXT, y TEXT, data JSONB)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT @extschema@.visualise(
    summary,
    @extschema@.bar_chart(
        summary,
        @extschema@.dataset(data),
        @extschema@.series(summary, x, y),
        @extschema@.option('fullHeight', true)
    ),
    @extschema@.option('header', false)
);
$$;


CREATE OR REPLACE FUNCTION @extschema@.quick_line_chart(summary TEXT, x TEXT, y TEXT, data JSONB)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT @extschema@.visualise(
    summary,
    @extschema@.line_chart(
        summary,
        @extschema@.dataset(data),
        @extschema@.series(summary, x, y),
        @extschema@.option('fullHeight', true)
    ),
    @extschema@.option('header', false)
);
$$;


CREATE OR REPLACE FUNCTION @extschema@.quick_scatter_chart(summary TEXT, x TEXT, y TEXT, data JSONB)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT @extschema@.visualise(
    summary,
    @extschema@.scatter_chart(
        summary,
        @extschema@.dataset(data),
        @extschema@.series(summary, x, y)
    ),
    @extschema@.option('header', false)
);
$$;


CREATE OR REPLACE FUNCTION @extschema@.quick_pie_chart(summary TEXT, x TEXT, y TEXT, data JSONB)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT @extschema@.visualise(
    summary,
    @extschema@.pie_chart(
        summary,
        @extschema@.dataset(data),
        @extschema@.series(summary, x, y),
        @extschema@.option('fullHeight', true)
    ),
    @extschema@.option('header', false)
);
$$;


CREATE OR REPLACE FUNCTION @extschema@.quick_polar_area_chart(summary TEXT, x TEXT, y TEXT, data JSONB)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT @extschema@.visualise(
    summary,
    @extschema@.polar_area_chart(
        summary,
        @extschema@.dataset(data),
        @extschema@.series(summary, x, y),
        @extschema@.option('fullHeight', true)
    ),
    @extschema@.option('header', false)
);
$$;


CREATE OR REPLACE FUNCTION @extschema@.quick_radar_chart(summary TEXT, x TEXT, y TEXT, data JSONB)
RETURNS @extschema@.t_definition LANGUAGE SQL AS $$
SELECT @extschema@.visualise(
    summary,
    @extschema@.radar_chart(
        summary,
        @extschema@.dataset(data),
        @extschema@.series(summary, x, y),
        @extschema@.option('fullHeight', true)
    ),
    @extschema@.option('header', false)
);
$$;
